import threading
import time

import flask
import gevent
import gevent.pywsgi
import geventwebsocket.handler
import pytest
import socketio
import socketio.exceptions
from locust_cloud.websocket import SessionMismatchError, Websocket, WebsocketTimeout

LOCUSTCLOUD_SESSION_ID = "valid-session-id"


@pytest.fixture(scope="session", autouse=True)
def websocket_server():
    fake_server_not_ready_app = flask.Flask("be_quiet")

    @fake_server_not_ready_app.route("/not-ready/")
    def not_ready():
        return "", 503

    sio = socketio.Server(async_handlers=True, always_connect=False, async_mode="gevent", cors_allowed_origins="*")
    sio_app = socketio.WSGIApp(sio, fake_server_not_ready_app)
    server = gevent.pywsgi.WSGIServer(
        ("", 1095), sio_app, log=None, error_log=None, handler_class=geventwebsocket.handler.WebSocketHandler
    )

    slow_reconnect = threading.Event()

    @sio.event
    def connect(sid, environ, auth):  # noqa: ARG001
        print("Client connected to websocket server")

        if auth != LOCUSTCLOUD_SESSION_ID:
            print("Rejected connection because it was for the wrong session")
            raise socketio.exceptions.ConnectionRefusedError("Session mismatch")

        if slow_reconnect.is_set():
            slow_reconnect.clear()
            sio.sleep(2)

    @sio.event
    def disconnect(sid):  # noqa: ARG001
        print("Client disconnected from websocket server")

    @sio.event
    def trigger_shutdown(sid, data):  # noqa: ARG001
        print("Got told by test to trigger shutdown")
        sio.call("events", {"events": [{"type": "shutdown", "message": None}], "id": 1}, to=sid, timeout=5)

    @sio.event
    def trigger_slow_reconnect(sid, data):  # noqa: ARG001
        print("Got told by test to trigger reconnect failure")
        slow_reconnect.set()
        sio.disconnect(sid)

    @sio.event
    def trigger_stderr(sid, data):  # noqa: ARG001
        print("Got told by test to send messages on stderr")
        sio.call("events", {"events": [{"type": "stderr", "message": data}], "id": 1}, to=sid, timeout=5)
        sio.call("events", {"events": [{"type": "stderr", "message": data}], "id": 2}, to=sid, timeout=5)

    @sio.event
    def trigger_duplicate_stderr(sid, data):  # noqa: ARG001
        print("Got told by test to send duplicate messages on stderr")
        sio.call("events", {"events": [{"type": "stderr", "message": data}], "id": 3}, to=sid, timeout=5)
        sio.call("events", {"events": [{"type": "stderr", "message": data}], "id": 3}, to=sid, timeout=5)

    def start_websocket_server():
        print("Starting websocket server")
        server.serve_forever()

    server_greenlet = gevent.spawn(start_websocket_server)

    try:
        time.sleep(1)
        yield
    finally:
        server.stop()
        server_greenlet.kill(block=True, timeout=None)


def test_websocket_server_sends_shutdown():
    ws = Websocket()
    ws.initial_connect_timeout = 2
    ws.connect(
        "http://127.0.0.1:1095",
        auth=LOCUSTCLOUD_SESSION_ID,
    )

    ws.sio.call("trigger_shutdown", "")
    ws.wait_timeout = 5
    assert ws.wait(timeout=True)


def test_websocket_server_not_available_before_timeout():
    ws = Websocket()
    ws.initial_connect_timeout = 2

    with pytest.raises(WebsocketTimeout) as exception:
        ws.connect(
            "http://127.0.0.1:1095/not-ready",
            auth=LOCUSTCLOUD_SESSION_ID,
        )

    assert str(exception.value) == "Timed out connecting to locust master"


def test_websocket_server_rejects_session():
    ws = Websocket()

    with pytest.raises(SessionMismatchError) as exception:
        ws.connect(
            "http://127.0.0.1:1095",
            auth="invalid-session-id",
        )

    assert str(exception.value) == "The session from this run of locust-cloud did not match the one on the server"


def test_websocket_deduplication(capsys):
    ws = Websocket()
    ws.connect(
        "http://127.0.0.1:1095",
        auth=LOCUSTCLOUD_SESSION_ID,
    )

    ws.sio.call("trigger_stderr", "banana\n")
    captured = capsys.readouterr()
    assert captured.err == "banana\nbanana\n"

    ws.sio.call("trigger_duplicate_stderr", "pineapple\n")

    captured = capsys.readouterr()
    assert captured.err == "pineapple\n"


def test_websocket_failed_reconnect():
    # FIXME: This test needs to be placed last. It messes up connecting from subsequent tests and I can't be bothered to debug it right now.
    ws = Websocket()
    ws.reconnect_timeout = 1
    ws.connect(
        "http://127.0.0.1:1095",
        auth=LOCUSTCLOUD_SESSION_ID,
    )

    ws.sio.emit("trigger_slow_reconnect", "")
    time.sleep(0.5)

    ws.wait_timeout = 5

    with pytest.raises(WebsocketTimeout) as exception:
        ws.wait(timeout=True)

    assert str(exception.value) == "Timed out connecting to locust master"



import logging
import sys
import threading
import urllib.parse

import socketio
import socketio.exceptions

logger = logging.getLogger(__name__)


class SessionMismatchError(Exception):
    pass


class WebsocketTimeout(Exception):
    pass


class Websocket:
    def __init__(self) -> None:
        """
        This class was created to encapsulate all the logic involved in the websocket implementation.
        The behaviour of the socketio client once a connection has been established
        is to try to reconnect forever if the connection is lost.
        The way this can be canceled is by setting the _reconnect_abort (threading.Event) on the client
        in which case it will simply proceed with shutting down without giving any indication of an error.
        This class handles timeouts for connection attempts as well as some logic around when the
        socket can be shut down. See descriptions on the methods for further details.
        """
        self.__shutdown_allowed = threading.Event()
        self.__timeout_on_disconnect = True
        self.initial_connect_timeout = 120
        self.reconnect_timeout = 10
        self.wait_timeout = 0
        self.exception: None | Exception = None

        self.sio = socketio.Client(handle_sigint=False)
        self.sio._reconnect_abort = threading.Event()
        # The _reconnect_abort value on the socketio client will be populated with a newly created threading.Event if it's not already set.
        # There is no way to set this by passing it in the constructor.
        # This event is the only way to interupt the retry logic when the connection is attempted.

        self.sio.on("connect", self.__on_connect)
        self.sio.on("disconnect", self.__on_disconnect)
        self.sio.on("connect_error", self.__on_connect_error)
        self.sio.on("events", self.__on_events)

        self.__processed_events: set[int] = set()

    def __set_connection_timeout(self, timeout) -> None:
        """
        Start a threading.Timer that will set the threading.Event on the socketio client
        that aborts any further attempts to reconnect, sets an exception on the websocket
        that will be raised from the wait method and the threading.Event __shutdown_allowed
        on the websocket that tells the wait method that it should stop blocking.
        """

        def _timeout():
            logger.debug(f"Websocket connection timed out after {timeout} seconds")
            self.sio._reconnect_abort.set()
            self.exception = WebsocketTimeout("Timed out connecting to locust master")
            self.__shutdown_allowed.set()

        self.__connect_timeout_timer = threading.Timer(timeout, _timeout)
        self.__connect_timeout_timer.daemon = True
        # logger.debug(f"Setting websocket connection timeout to {timeout} seconds")
        self.__connect_timeout_timer.start()

    def connect(self, url, *, auth) -> None:
        """
        Send along retry=True when initiating the socketio client connection
        to make it use it's builtin logic for retrying failed connections that
        is usually used for reconnections. This will retry forever.
        When connecting start a timer to trigger disabling the retry logic and
        raise a WebsocketTimeout exception.
        """
        ws_connection_info = urllib.parse.urlparse(url)
        self.__set_connection_timeout(self.initial_connect_timeout)
        try:
            self.sio.connect(
                f"{ws_connection_info.scheme}://{ws_connection_info.netloc}",
                auth=auth,
                retry=True,
                **{"socketio_path": ws_connection_info.path} if ws_connection_info.path else {},
            )
        except socketio.exceptions.ConnectionError:
            if self.exception:
                raise self.exception

            raise

    def shutdown(self) -> None:
        """
        When shutting down the socketio client a disconnect event will fire.
        Before doing so disable the behaviour of starting a threading.Timer
        to handle timeouts on attempts to reconnect since no further such attempts
        will be made.
        If such a timer is already running, cancel it since the client is being shutdown.
        """
        self.__timeout_on_disconnect = False
        if hasattr(self, "__connect_timeout_timer"):
            self.__connect_timeout_timer.cancel()
        self.sio.shutdown()

    def wait(self, timeout=False) -> bool:
        """
        Block until the threading.Event __shutdown_allowed is set, with a timeout if indicated.
        If an exception has been set on the websocket (from a connection timeout timer or the
        __on_connect_error method), raise it.
        """
        timeout = self.wait_timeout if timeout else None
        if timeout:  # not worth even debug logging if we dont have a timeout
            logger.debug(f"Waiting for shutdown for {str(timeout) + 's' if timeout else 'ever'}")
        res = self.__shutdown_allowed.wait(timeout)
        if self.exception:
            raise self.exception
        return res

    def __on_connect(self) -> None:
        """
        This gets events whenever a connection is successfully established.
        When this happens, cancel the running threading.Timer that would
        abort reconnect attempts and raise a WebsocketTimeout exception.
        The wait_timeout is originally set to zero when creating the websocket
        but once a connection has been established this is raised to ensure
        that the server is given the chance to send all the logs and an
        official shutdown event.
        """
        self.__connect_timeout_timer.cancel()
        self.wait_timeout = 90
        logger.debug("Websocket connected")

    def __on_disconnect(self) -> None:
        """
        This gets events whenever a connection is lost.
        The socketio client will try to reconnect forever so,
        unless the behaviour has been disabled, a threading.Timer
        is started that will abort reconnect attempts and raise a
        WebsocketTimeout exception.
        """
        if self.__timeout_on_disconnect:
            self.__set_connection_timeout(self.reconnect_timeout)
        logger.debug("Websocket disconnected")

    def __on_events(self, data):
        """
        This gets events explicitly sent by the websocket server.
        This will either be messages to print on stdout/stderr or
        an indication that the CLI can shut down in which case the
        threading.Event __shutdown_allowed gets set on the websocket
        that tells the wait method that it should stop blocking.
        """
        shutdown = False
        shutdown_message = ""

        if data["id"] in self.__processed_events:
            logger.debug(f"Got duplicate data on websocket, id {data['id']}")
            return

        self.__processed_events.add(data["id"])

        for event in data["events"]:
            type = event["type"]

            if type == "shutdown":
                shutdown = True
                shutdown_message = event["message"]
            elif type == "stdout":
                sys.stdout.write(event["message"])
            elif type == "stderr":
                sys.stderr.write(event["message"])
            else:
                raise Exception("Unexpected event type")

        if shutdown:
            logger.debug("Got shutdown from locust master")
            self.__connect_timeout_timer.cancel()  # I dont know exactly why/if this is necessary but we had an issue in status-checker once
            if shutdown_message:
                print(shutdown_message)

            self.__shutdown_allowed.set()

    def __on_connect_error(self, data) -> None:
        """
        This gets events whenever there's an error during connection attempts.
        The specific case that is handled below is triggered when the connection
        is made with the auth parameter not matching the session ID on the server.
        If this error occurs it's because the connection is attempted towards an
        instance of locust not started by this CLI.

        In that case:
        Cancel the running threading.Timer that would abort reconnect attempts
        and raise a WebsocketTimeout exception.
        Set an exception on the websocket that will be raised from the wait method.
        Cancel further reconnect attempts.
        Set the threading.Event __shutdown_allowed on the websocket that tells the
        wait method that it should stop blocking.
        """
        # Do nothing if it's not the specific case we know how to deal with
        if not (isinstance(data, dict) and data.get("message") == "Session mismatch"):
            return

        self.__connect_timeout_timer.cancel()
        self.exception = SessionMismatchError(
            "The session from this run of locust-cloud did not match the one on the server"
        )
        self.sio._reconnect_abort.set()
        self.__shutdown_allowed.set()



import sys
import time
import webbrowser

import requests
from locust_cloud.common import VALID_REGIONS, CloudConfig, delete_cloud_config, get_api_url, write_cloud_config

POLLING_FREQUENCY = 1


def web_login() -> None:
    print("Enter the number for the region to authenticate against")
    print()
    for i, valid_region in enumerate(VALID_REGIONS, start=1):
        print(f"  {i}. {valid_region}")
    print()
    choice = input("> ")
    try:
        region_index = int(choice) - 1
        assert 0 <= region_index < len(VALID_REGIONS)
    except (ValueError, AssertionError):
        print(f"Not a valid choice: '{choice}'")
        sys.exit(1)

    region = VALID_REGIONS[region_index]

    try:
        response = requests.post(f"{get_api_url(region)}/cli-auth")
        response.raise_for_status()
        response_data = response.json()
        authentication_url = response_data["authentication_url"]
        result_url = response_data["result_url"]
    except Exception as e:
        print("Something went wrong trying to authorize the locust-cloud CLI:", str(e))
        sys.exit(1)

    message = f"""
Attempting to automatically open the SSO authorization page in your default browser.
If the browser does not open or you wish to use a different device to authorize this request, open the following URL:

{authentication_url}
    """.strip()
    print()
    print(message)

    webbrowser.open_new_tab(authentication_url)

    while True:  # Should there be some kind of timeout?
        response = requests.get(result_url)

        if not response.ok:
            print("Oh no!")
            print(response.text)
            sys.exit(1)

        data = response.json()

        if data["state"] == "pending":
            time.sleep(POLLING_FREQUENCY)
            continue
        elif data["state"] == "failed":
            print(f"\nFailed to authorize CLI: {data['reason']}")
            sys.exit(1)
        elif data["state"] == "authorized":
            print("\nAuthorization succeded. Now you can start a cloud run using: locust --cloud ...")
            break
        else:
            print("\nGot unexpected response when authorizing CLI")
            sys.exit(1)

    config = CloudConfig(
        id_token=data["id_token"],
        refresh_token=data["refresh_token"],
        user_sub_id=data["user_sub_id"],
        refresh_token_expires=data["refresh_token_expires"],
        id_token_expires=data["id_token_expires"],
        region=region,
    )
    write_cloud_config(config)


def logout():
    delete_cloud_config()



import argparse
import base64
import gzip
import io
import os
import pathlib
import shutil
import sys
import tempfile

from locust_cloud.actions import delete
from locust_cloud.apisession import ApiSession
from locust_cloud.web_login import logout, web_login

if sys.version_info >= (3, 11):
    import tomllib
else:
    import tomli as tomllib

from argparse import ArgumentTypeError
from collections import OrderedDict
from collections.abc import Callable, Generator
from typing import IO, Any, cast
from zipfile import ZipFile

import configargparse

CWD = pathlib.Path.cwd()


class LocustTomlConfigParser(configargparse.TomlConfigParser):
    def parse(self, stream: IO[str]) -> OrderedDict[str, Any]:
        try:
            config = tomllib.loads(stream.read())
        except Exception as e:
            raise configargparse.ConfigFileParserException(f"Couldn't parse TOML file: {e}")

        result: OrderedDict[str, Any] = OrderedDict()

        for section in self.sections:
            data = configargparse.get_toml_section(config, section)
            if data:
                for key, value in data.items():
                    if isinstance(value, list):
                        result[key] = value
                    elif value is not None:
                        result[key] = str(value)
                break

        return result


def pipe(value: Any, *functions: Callable) -> Any:
    for function in functions:
        value = function(value)

    return value


def valid_extra_files_path(file_path: str) -> pathlib.Path:
    p = pathlib.Path(file_path).resolve()

    if not CWD in p.parents:
        raise ArgumentTypeError(f"Can only reference files under current working directory: {CWD}")
    if not p.exists():
        raise ArgumentTypeError(f"File not found: {file_path}")
    return p


def valid_extra_packages_path(file_path: str) -> pathlib.Path:
    p = pathlib.Path(file_path).resolve()

    if not p.exists():
        raise ArgumentTypeError(f"Path not found: {file_path}")
    if p.is_file() and not (p.suffix == ".whl" or p.suffixes == [".tar", ".gz"]):
        raise ArgumentTypeError(f"Invalid file suffix (must be '.whl' or '.tar.gz'): {file_path}")

    return p


def transfer_encode(file_name: str, stream: IO[bytes]) -> dict[str, str]:
    return {
        "filename": file_name,
        "data": pipe(
            stream.read(),
            gzip.compress,
            base64.b64encode,
            bytes.decode,
        ),
    }


def transfer_encoded_file(file_path: str) -> dict[str, str]:
    try:
        with open(file_path, "rb") as f:
            return transfer_encode(os.path.basename(file_path), f)
    except FileNotFoundError:
        raise ArgumentTypeError(f"File not found: {file_path}")


def expanded(paths: list[pathlib.Path]) -> Generator[pathlib.Path, None, None]:
    for path in paths:
        if path.is_dir():
            for root, _, file_names in os.walk(path):
                for file_name in file_names:
                    yield pathlib.Path(root) / file_name
        else:
            yield path


def transfer_encoded_args_files(paths: list[pathlib.Path], to_file: str | None) -> dict[str, str]:
    buffer = io.BytesIO()

    with ZipFile(buffer, "w") as zf:
        for path in set(expanded(paths)):
            zf.write(path.relative_to(CWD))

    buffer.seek(0)
    return transfer_encode(f"{to_file}.zip", buffer)


def flat_transfer_encoded_args_files(paths: list[pathlib.Path], to_file: str | None) -> dict[str, str]:
    buffer = io.BytesIO()

    with tempfile.TemporaryDirectory() as tmpdir:
        tmp_path = pathlib.Path(tmpdir)

        for src in paths:
            src_path = pathlib.Path(src)
            dest_path = tmp_path / src_path.name

            if src_path.is_file():
                shutil.copy(src_path, dest_path)
            elif src_path.is_dir():
                shutil.copytree(src_path, dest_path)
            else:
                print(f"Warning: {src} is not a valid file or directory")

        # Create the zip archive
        with ZipFile(buffer, "w") as zf:
            for item in tmp_path.iterdir():
                if item.is_file():
                    zf.write(item, arcname=item.name)
                elif item.is_dir():
                    for root, _, files in os.walk(item):
                        for file in files:
                            file_path = pathlib.Path(root) / file
                            arcname = file_path.relative_to(tmp_path)
                            zf.write(file_path, arcname)

    buffer.seek(0)
    return transfer_encode(f"{to_file}.zip", buffer)


class MergeToTransferEncodedZip(argparse.Action):
    def __call__(self, parser, namespace, values, option_string=None):
        paths = cast(list[pathlib.Path], values)
        value = transfer_encoded_args_files(paths, option_string.lstrip("-"))
        setattr(namespace, self.dest, value)


class MergeToTransferEncodedZipFlat(MergeToTransferEncodedZip):
    def __call__(self, parser, namespace, values, option_string=None):
        paths = cast(list[pathlib.Path], values)
        value = flat_transfer_encoded_args_files(paths, option_string.lstrip("-"))
        setattr(namespace, self.dest, value)


class WebLogin(argparse.Action):
    def __call__(self, parser, namespace, values, option_string=None):
        web_login()
        parser.exit()


class WebLogout(argparse.Action):
    def __call__(self, parser, namespace, values, option_string=None):
        logout()
        parser.exit()


class StackTeardown(argparse.Action):
    def __call__(self, parser, namespace, values, option_string=None):
        session = ApiSession(namespace.non_interactive)
        delete(session)
        parser.exit()


cloud_parser = configargparse.ArgumentParser(add_help=False)
cloud_parser.add_argument(
    "--login",
    nargs=0,
    action=WebLogin,
    help="Launch an interactive session to authenticate your user.\nOnce completed your credentials will be stored and automatically refreshed for quite a long time.\nOnce those expire you will be prompted to perform another login.",
)
cloud_parser.add_argument(
    "--logout",
    nargs=0,
    action=WebLogout,
    help="Removes the authentication credentials",
)
cloud_parser.add_argument(
    "--delete",
    nargs=0,
    action=StackTeardown,
    help="Delete a running cluster. Useful if locust-cloud was killed/disconnected or if there was an error.",
)
cloud_parser.add_argument(
    "--requirements",
    metavar="<filename>",
    type=transfer_encoded_file,
    help="Optional requirements.txt file that contains your external libraries.",
)
cloud_parser.add_argument(
    "--non-interactive",
    action="store_true",
    default=False,
    help="This can be set when, for example, running in a CI/CD environment to ensure no interactive steps while executing.\nRequires that LOCUSTCLOUD_USERNAME, LOCUSTCLOUD_PASSWORD and LOCUSTCLOUD_REGION environment variables are set.",
    env_var="LOCUSTCLOUD_NON_INTERACTIVE",
)
cloud_parser.add_argument(
    "--workers",
    metavar="<int>",
    type=int,
    help="Number of workers to use for the deployment. Defaults to number of users divided by 500, but the default may be customized for your account.",
    default=None,
)
cloud_parser.add_argument(
    "--image-tag",
    type=str,
    default=None,
    help=configargparse.SUPPRESS,  # overrides the locust-cloud docker image tag. for internal use
)
cloud_parser.add_argument(
    "--mock-server",
    action="store_true",
    default=False,
    help="Start a demo mock service and set --host parameter to point Locust towards it.",
)
cloud_parser.add_argument(
    "--extra-files",
    action=MergeToTransferEncodedZip,
    nargs="*",
    type=valid_extra_files_path,
    help="A list of extra files or directories to upload. Space-separated, e.g. `--extra-files testdata.csv *.py my-directory/`.",
)
cloud_parser.add_argument(
    "--extra-packages",
    action=MergeToTransferEncodedZipFlat,
    nargs="*",
    type=valid_extra_packages_path,
    help="A list of extra packages to upload. Space-separated whl/tar.gz files or directory packages to be installed when running locust.",
)
cloud_parser.add_argument(
    "--testrun-tags",
    nargs="*",
    default=None,
    help="A list of tags that can be used to filter testruns.",
)

combined_cloud_parser = configargparse.ArgumentParser(
    parents=[cloud_parser],
    default_config_files=[
        "~/.cloud.conf",
        "cloud.conf",
    ],
    auto_env_var_prefix="LOCUSTCLOUD_",
    formatter_class=configargparse.RawTextHelpFormatter,
    config_file_parser_class=configargparse.CompositeConfigParser(
        [
            LocustTomlConfigParser(["tool.locust"]),
            configargparse.DefaultConfigFileParser,
        ]
    ),
    description="""Launches a distributed Locust runs on locust.cloud infrastructure.

Example: locust --cloud -f my_locustfile.py --users 1000 ...""",
    epilog="""Any parameters not listed here are forwarded to locust master unmodified, so go ahead and use things like --users, --host, --run-time, ...
Locust config can also be set using config file (~/.locust.conf, locust.conf, pyproject.toml, ~/.cloud.conf or cloud.conf).
Parameters specified on command line override env vars, which in turn override config files.""",
    add_config_file_help=False,
    add_env_var_help=False,
)
combined_cloud_parser.add_argument(
    "-f",
    "--locustfile",
    metavar="<filename>",
    default="locustfile.py",
    help="The Python file that contains your test. Defaults to 'locustfile.py'.",
    env_var="LOCUST_LOCUSTFILE",
    type=transfer_encoded_file,
)
combined_cloud_parser.add_argument(
    "-u",
    "--users",
    type=int,
    default=1,
    help="Number of users to launch. This is the same as the regular Locust argument, but also affects how many workers to launch.",
    env_var="LOCUST_USERS",
)
combined_cloud_parser.add_argument(
    "--loglevel",
    "-L",
    type=str.upper,
    help="Set --loglevel DEBUG for extra info.",
    choices=["DEBUG", "INFO", "WARNING", "ERROR"],
    default="INFO",
)


def add_locust_cloud_argparse(parser):
    cloud_group = parser.add_argument_group(
        "Locust Cloud",
        """Launches a distributed Locust run on locust.cloud infrastructure.

Example: locust --cloud -f my_locustfile.py --users 1000 ...""",
    )

    # This arguments is defined here because only makes sense when
    # running from locust core
    cloud_group.add_argument(
        "--cloud",
        action="store_true",
        help="Run Locust in cloud mode.",
    )

    for action in cloud_parser._actions:
        cloud_group._add_action(action)



import logging
import os
import time
import webbrowser
from threading import Thread

import requests
from locust_cloud.actions import delete
from locust_cloud.apisession import ApiSession
from locust_cloud.args import combined_cloud_parser
from locust_cloud.common import __version__
from locust_cloud.input_events import input_listener
from locust_cloud.web_login import logout, web_login
from locust_cloud.websocket import SessionMismatchError, Websocket, WebsocketTimeout

logger = logging.getLogger(__name__)


def configure_logging(loglevel: str) -> None:
    format = (
        "[%(asctime)s] %(levelname)s: %(message)s"
        if loglevel == "INFO"
        else "[%(asctime)s] %(levelname)s/%(module)s: %(message)s"
    )
    logging.basicConfig(format=format, level=loglevel)
    # Restore log level for other libs. Yes, this can probably be done more nicely
    logging.getLogger("requests").setLevel(logging.INFO)
    logging.getLogger("urllib3").setLevel(logging.INFO)


def main():
    options, locust_options = combined_cloud_parser.parse_known_args()

    configure_logging(options.loglevel)

    if not options.locustfile:
        logger.error("A locustfile is required to run a test.")
        return 1

    if options.login:
        try:
            web_login()
        except Exception:
            pass
        return
    if options.logout:
        try:
            logout()
        except Exception:
            pass
        return

    session = ApiSession(options.non_interactive)
    websocket = Websocket()

    if options.delete:
        delete(session)
        return

    try:
        logger.info(f"Deploying ({session.region}, {__version__})")
        locust_env_variables = [
            {"name": env_variable, "value": os.environ[env_variable]}
            for env_variable in os.environ
            if env_variable.startswith("LOCUST_")
            and env_variable
            not in [
                "LOCUST_LOCUSTFILE",
                "LOCUST_USERS",
                "LOCUST_WEB_HOST_DISPLAY_NAME",
                "LOCUST_SKIP_MONKEY_PATCH",
                "LOCUST_CLOUD",
            ]
        ]

        locust_args = [
            {"name": "LOCUST_USERS", "value": str(options.users)},
            {"name": "LOCUST_FLAGS", "value": " ".join([option for option in locust_options if option != "--cloud"])},
            {"name": "LOCUSTCLOUD_DEPLOYER_URL", "value": session.api_url},
            *locust_env_variables,
        ]

        if options.testrun_tags:
            locust_args.append({"name": "LOCUSTCLOUD_TESTRUN_TAGS", "value": ",".join(options.testrun_tags)})

        payload = {
            "locust_args": locust_args,
            "locustfile": options.locustfile,
            "user_count": options.users,
            "mock_server": options.mock_server,
        }

        if options.image_tag is not None:
            payload["image_tag"] = options.image_tag

        if options.workers is not None:
            payload["worker_count"] = options.workers

        if options.requirements:
            payload["requirements"] = options.requirements

        if options.extra_files:
            payload["extra_files"] = options.extra_files

        if options.extra_packages:
            payload["extra_packages"] = options.extra_packages

        for attempt in range(1, 16):
            try:
                response = session.post("/deploy", json=payload)
                js = response.json()

                if response.status_code != 202:
                    # 202 means the stack is currently terminating, so we retry
                    break

                if attempt == 1:
                    logger.info(js["message"])

                time.sleep(2)
            except requests.exceptions.RequestException as e:
                logger.error(f"Failed to deploy the load generators: {e}")
                return 1
        else:
            logger.error("Your Locust instance is still running, run locust --delete")
            return 1

        if response.status_code != 200:
            try:
                logger.error(f"{js['Message']} (HTTP {response.status_code}/{response.reason})")
            except Exception:
                logger.error(
                    f"HTTP {response.status_code}/{response.reason} - Response: {response.text} - URL: {response.request.url}"
                )
            return 1

        log_ws_url = js["log_ws_url"]
        session_id = js["session_id"]
        webui_url = log_ws_url.replace("/socket-logs", "")

        def open_ui():
            webbrowser.open_new_tab(webui_url)

        Thread(target=input_listener({"\r": open_ui, "\n": open_ui}), daemon=True).start()

        # logger.debug(f"Session ID is {session_id}")

        logger.info("Waiting for load generators to be ready...")
        websocket.connect(
            log_ws_url,
            auth=session_id,
        )
        websocket.wait()

    except KeyboardInterrupt:
        logger.debug("Interrupted by user")
        delete(session)
        try:
            websocket.wait(timeout=True)
        except (WebsocketTimeout, SessionMismatchError) as e:
            logger.error(str(e))
            return 1
    except WebsocketTimeout as e:
        logger.error(str(e))
        delete(session)
        return 1
    except SessionMismatchError as e:
        # In this case we do not trigger the teardown since the running instance is not ours
        logger.error(str(e))
        return 1
    except Exception as e:
        logger.exception(e)
        delete(session)
        return 1
    else:
        delete(session)
    finally:
        logger.debug("Shutting down websocket")
        websocket.shutdown()



import base64
import gzip
import io
import pathlib
from argparse import ArgumentTypeError
from zipfile import ZipFile

import pytest
from locust_cloud.args import (
    CWD,
    combined_cloud_parser,
    expanded,
    pipe,
    transfer_encode,
    transfer_encoded_args_files,
    transfer_encoded_file,
    valid_extra_files_path,
)


def test_pipe():
    one = lambda x: x * 3
    two = lambda x: x + 3
    assert pipe(4, one, two) == 15


def test_valid_extra_files_path():
    with pytest.raises(ArgumentTypeError) as exception:
        valid_extra_files_path("../pineapple")

    assert str(exception.value) == f"Can only reference files under current working directory: {CWD}"

    bad_path = str(CWD / "does-not-exist")

    with pytest.raises(ArgumentTypeError) as exception:
        valid_extra_files_path(bad_path)

    assert str(exception.value) == f"File not found: {bad_path}"


def test_transfer_encode():
    file_name = "pineapple.txt"
    data = b"pineapple"
    result = transfer_encode(file_name, io.BytesIO(data))
    assert file_name == result["filename"]
    assert data == pipe(
        result["data"],
        str.encode,
        base64.b64decode,
        gzip.decompress,
    )


def test_transfer_encoded_file():
    with pytest.raises(ArgumentTypeError) as exception:
        transfer_encoded_file("does-not-exist")

    assert str(exception.value) == "File not found: does-not-exist"


def test_expanded():
    result = list(expanded([pathlib.Path("locustfile.py"), pathlib.Path("testdata/extra-files")]))
    assert result == [pathlib.Path("locustfile.py"), pathlib.Path("testdata/extra-files/extra.txt")]


def test_transfer_encoded_args_files():
    result = transfer_encoded_args_files([pathlib.Path("testdata/extra-files").resolve()], "extra-files")
    assert result["filename"] == "extra-files.zip"

    buffer = pipe(
        result["data"],
        str.encode,
        base64.b64decode,
        gzip.decompress,
        io.BytesIO,
    )

    with ZipFile(buffer) as zf:
        assert zf.namelist() == ["testdata/extra-files/extra.txt"]


def test_parser_locustfile(capsys):
    with pytest.raises(SystemExit):
        combined_cloud_parser.parse_known_args("locust-cloud --locustfile does-not-exist")

    expected = "error: argument -f/--locustfile: File not found: does-not-exist"
    assert expected in capsys.readouterr().err

    options, _ = combined_cloud_parser.parse_known_args("locust-cloud --locustfile testdata/extra-files/extra.txt")
    assert options.locustfile == transfer_encoded_file("testdata/extra-files/extra.txt")


def test_parser_extra_files(capsys):
    with pytest.raises(SystemExit):
        combined_cloud_parser.parse_known_args("locust-cloud --extra-files ../pineapple")

    expected = "error: argument --extra-files: Can only reference files under current working directory"
    assert expected in capsys.readouterr().err

    with pytest.raises(SystemExit):
        combined_cloud_parser.parse_known_args("locust-cloud --extra-files does-not-exist")

    expected = "error: argument --extra-files: File not found: does-not-exist"
    assert expected in capsys.readouterr().err

    options, _ = combined_cloud_parser.parse_known_args("locust-cloud --extra-files testdata/extra-files")
    assert options.extra_files["filename"] == "extra-files.zip"
    buffer = pipe(
        options.extra_files["data"],
        str.encode,
        base64.b64decode,
        gzip.decompress,
        io.BytesIO,
    )

    with ZipFile(buffer) as zf:
        assert zf.namelist() == ["testdata/extra-files/extra.txt"]


def test_parser_loglevel(capsys):
    options, _ = combined_cloud_parser.parse_known_args("locust-cloud --loglevel DEBUG")
    assert options.loglevel == "DEBUG"

    with pytest.raises(SystemExit):
        combined_cloud_parser.parse_known_args("locust-cloud --loglevel pineapple")

    expected = "error: argument --loglevel/-L: invalid choice: 'PINEAPPLE'"
    assert expected in capsys.readouterr().err



import logging

logger = logging.getLogger(__name__)


def delete(session):
    try:
        logger.info("Tearing down Locust cloud...")
        response = session.delete(
            "/teardown",
        )

        if response.status_code == 200:
            logger.debug(f"Response message from teardown: {response.json()['message']}")
        else:
            logger.info(
                f"Could not automatically tear down Locust Cloud: HTTP {response.status_code}/{response.reason} - Response: {response.text} - URL: {response.request.url}"
            )
    except KeyboardInterrupt:
        pass  # don't show nasty callstack
    except Exception as e:
        logger.error(f"Could not automatically tear down Locust Cloud: {e.__class__.__name__}:{e}")



import builtins
import webbrowser
from unittest.mock import MagicMock

import locust_cloud.common
import locust_cloud.web_login
import pytest
import requests_mock

AUTH_ID = "fake"
REGION = "eu-north-1"
API_URL = locust_cloud.common.get_api_url(REGION)
REGION_CHOICE = locust_cloud.common.VALID_REGIONS.index(REGION) + 1


@pytest.fixture
def mocked_requests():
    with requests_mock.Mocker() as m:
        m.post(
            f"{API_URL}/cli-auth",
            json={
                "authentication_url": "<url goes here>",
                "result_url": f"{API_URL}/cli-auth/result/{AUTH_ID}",
            },
        )
        yield m


@pytest.fixture
def select_region(monkeypatch):
    monkeypatch.setattr(builtins, "input", lambda prompt: REGION_CHOICE)


@pytest.fixture(autouse=True)
def faster_cli_auth_result_polling(monkeypatch):
    monkeypatch.setattr(locust_cloud.web_login, "POLLING_FREQUENCY", 0.1)


@pytest.fixture(autouse=True)
def block_browser_launch(monkeypatch):
    monkeypatch.setattr(webbrowser, "open_new_tab", lambda url: None)


def test_browser_login_failed(mocked_requests, select_region, capsys):  # noqa: ARG001
    mocked_requests.get(
        f"{API_URL}/cli-auth/result/{AUTH_ID}",
        [{"json": {"state": "pending"}}, {"json": {"state": "failed", "reason": "because"}}],
    )
    with pytest.raises(SystemExit):
        locust_cloud.web_login.web_login()

    expected = "Failed to authorize CLI: because"
    assert expected in capsys.readouterr().out


def test_browser_login_succeded(mocked_requests, select_region, monkeypatch, capsys):  # noqa: ARG001
    mock = MagicMock()
    monkeypatch.setattr(locust_cloud.web_login, "write_cloud_config", mock)

    response = {
        "state": "authorized",
        "id_token": "A",
        "refresh_token": "B",
        "user_sub_id": "C",
        "refresh_token_expires": 42,
        "id_token_expires": 52,
    }

    mocked_requests.get(
        f"{API_URL}/cli-auth/result/{AUTH_ID}",
        [{"json": {"state": "pending"}}, {"json": response}],
    )
    locust_cloud.web_login.web_login()

    expected = "Authorization succeded. Now you can start a cloud run using: locust --cloud ..."
    assert expected in capsys.readouterr().out

    expected_cloud_config = locust_cloud.common.CloudConfig(
        id_token=response["id_token"],
        refresh_token=response["refresh_token"],
        user_sub_id=response["user_sub_id"],
        refresh_token_expires=response["refresh_token_expires"],
        id_token_expires=response["id_token_expires"],
        region=REGION,
    )
    mock.assert_called_once_with(expected_cloud_config)



import json
import os
import pathlib
import re
import select
import subprocess
import time
from datetime import timedelta
from urllib.parse import parse_qs, urlparse

import locust_cloud.common
import platformdirs
import pytest
import requests
from locust_cloud.apisession import unauthorized_message

CLOUD_CONFIG_FILE = pathlib.Path(platformdirs.user_config_dir(appname="locust-cloud")) / "config"
LOCUSTCLOUD_USERNAME = os.environ["LOCUSTCLOUD_USERNAME"]
LOCUSTCLOUD_PASSWORD = os.environ["LOCUSTCLOUD_PASSWORD"]
REGION = "eu-north-1"
API_URL = locust_cloud.common.get_api_url(REGION)


@pytest.fixture(scope="session", autouse=True)
def backup_cloud_config_file():
    do_backup = CLOUD_CONFIG_FILE.exists()
    backup_file = CLOUD_CONFIG_FILE.parent / (CLOUD_CONFIG_FILE.name + "_backup")

    if do_backup:
        CLOUD_CONFIG_FILE.rename(backup_file)

    try:
        yield
    finally:
        if do_backup:
            backup_file.rename(CLOUD_CONFIG_FILE)

        else:  # was not there before the test, should not be there after
            if CLOUD_CONFIG_FILE.exists():
                CLOUD_CONFIG_FILE.unlink()


def check_for_output(stream, regex, timeout=None) -> re.Match | None:
    start = time.time()

    while True:
        res = select.select([stream], [], [], 1.0)

        if res[0]:
            line = stream.readline()
            if m := regex.match(line):
                return m

        if timeout and time.time() - start > timeout:
            break

    return None


def test_cli_auth() -> None:
    # Ensure there is no cloud config file by deleting it
    if CLOUD_CONFIG_FILE.exists():
        CLOUD_CONFIG_FILE.unlink()
    assert not CLOUD_CONFIG_FILE.exists()

    env = {
        "PATH": os.environ["PATH"],  # Needed for obvious reasons
        "PYTHONUNBUFFERED": "1",  # To avoid issues when reading from the process output stream
        "BROWSER": "echo 'Fake browser launch %s'",  # Prevents the automatic opening of a browser window during the test
    }

    # Check that we get the message that login is required
    process = subprocess.Popen(["locust-cloud"], env=env, stdout=subprocess.PIPE, stderr=subprocess.PIPE, text=True)
    process.wait(timeout=1)
    assert process.stdout  # typing, not testing...
    assert unauthorized_message in process.stdout.read()

    # Do a locust-cloud --login
    process = subprocess.Popen(
        ["locust-cloud", "--login"],
        env=env,
        stdout=subprocess.PIPE,
        stdin=subprocess.PIPE,
        text=True,
    )
    assert process.stdout  # typing, not testing...
    assert process.stdin  # typing, not testing...

    # Provide input for the choosing of the region
    process.stdin.write("2\n")
    process.stdin.flush()

    # Wait for the login url to be printed
    m = check_for_output(process.stdout, re.compile(r"(http.*)"), timeout=10)
    assert m, "Didn't get a url in the output before timeout"
    url = m.groups()[0]

    # Make a request to the login url
    session = requests.Session()
    response = requests.get(url)
    assert response.ok, "Failed to load login url"

    auth_id = parse_qs(urlparse(url).query).get("auth_id", [])[0]

    assert auth_id, "Auth id is missing"

    # Login form submits to the deployer, so we'll skip a step
    response = session.post(
        f"{API_URL}/authenticate",
        json={"email": LOCUSTCLOUD_USERNAME, "password": LOCUSTCLOUD_PASSWORD, "auth_id": auth_id},
    )

    # Check that we end up on the success page
    assert response.json().get("auth_id") == auth_id

    # Wait for the process to finish
    process.wait()

    # Check that a cloud config file has been created
    assert CLOUD_CONFIG_FILE.exists()
    cloud_config = json.loads(CLOUD_CONFIG_FILE.read_text())
    for key in ("id_token", "refresh_token", "refresh_token_expires", "region"):
        assert key in cloud_config, f"Missing cloud config key: '{key}'"

    # Check that refresh_token_expires is around 365 days in the future
    expires_in = cloud_config["refresh_token_expires"] - time.time()
    assert -10 < expires_in - timedelta(days=365).total_seconds() < 10



# Simple locustfile for testing against example target
import pathlib
import random

from locust import FastHttpUser, run_single_user, task  # pyright: ignore [reportMissingImports]

product_ids = [1, 2, 42, 4711]


class MyUser(FastHttpUser):
    @task
    def t(self) -> None:
        with self.rest("POST", "/authenticate", json={"username": "foo", "password": "bar"}) as resp:
            if error := resp.js.get("error"):
                resp.failure(error)

        for product_id in random.sample(product_ids, 2):
            with self.rest("POST", "/cart/add", json={"productId": product_id}) as resp:
                pass

        with self.rest("POST", "/checkout/confirm") as resp:
            if not resp.js.get("orderId"):
                resp.failure("orderId missing")


extra = pathlib.Path("testdata/extra-files/extra.txt")
if extra.exists():
    print("--extra-files verification:", extra.read_text())

try:
    import example  # type: ignore

    example.hello()
except ImportError:
    pass  # ignore this for local runs

try:
    import dotenv  # type: ignore # noqa: F401

    print("dotenv imported successfully, --requirements seems to be working")
except ImportError:
    pass  # ignore this for local runs


if __name__ == "__main__":
    run_single_user(MyUser)



import logging
import os
import sys
import time

import requests
from locust_cloud.common import VALID_REGIONS, __version__, get_api_url, read_cloud_config, write_cloud_config

logger = logging.getLogger(__name__)

unauthorized_message = "You need to log in again. Please run:\n    locust --login"


class ApiSession(requests.Session):
    def __init__(self, non_interactive: bool) -> None:
        super().__init__()
        self.non_interactive = non_interactive

        if non_interactive:
            username = os.getenv("LOCUSTCLOUD_USERNAME")
            password = os.getenv("LOCUSTCLOUD_PASSWORD")
            region = os.getenv("LOCUSTCLOUD_REGION")

            if not all([username, password, region]):
                print(
                    "Running with --non-interactive requires that LOCUSTCLOUD_USERNAME, LOCUSTCLOUD_PASSWORD and LOCUSTCLOUD_REGION environment variables are set."
                )
                sys.exit(1)

            if region not in VALID_REGIONS:
                print("Environment variable LOCUSTCLOUD_REGION needs to be set to one of", ", ".join(VALID_REGIONS))
                sys.exit(1)

            self.__configure_for_region(region)
            response = requests.post(
                self.__login_url,
                json={"username": username, "password": password},
                headers={"X-Client-Version": __version__},
            )
            if not response.ok:
                print(f"Authentication failed: {response.text}")
                sys.exit(1)

            id_token = response.json()["cognito_client_id_token"]
            user_sub_id = response.json()["user_sub_id"]
            refresh_token = response.json()["refresh_token"]
            id_token_expires = response.json()["id_token_expires"]
        else:
            config = read_cloud_config()

            if config.refresh_token_expires < time.time() + 24 * 60 * 60:
                print(unauthorized_message)
                sys.exit(1)

            assert config.region
            self.__configure_for_region(config.region)
            id_token = config.id_token
            user_sub_id = config.user_sub_id
            refresh_token = config.refresh_token
            id_token_expires = config.id_token_expires

        assert id_token

        self.__user_sub_id = user_sub_id
        self.__refresh_token = refresh_token
        self.__id_token_expires = id_token_expires - 60  # Refresh 1 minute before expiry
        self.headers["Authorization"] = f"Bearer {id_token}"
        self.headers["X-Client-Version"] = __version__

    def __configure_for_region(self, region: str) -> None:
        self.region = region
        self.api_url = get_api_url(region)
        self.__login_url = f"{self.api_url}/auth/login"

        logger.debug(f"Lambda url: {self.api_url}")

    def __ensure_valid_authorization_header(self) -> None:
        if self.__id_token_expires > time.time():
            return
        if not self.__user_sub_id and self.__refresh_token:
            print(unauthorized_message)
            sys.exit(1)

        response = requests.post(
            self.__login_url,
            json={"user_sub_id": self.__user_sub_id, "refresh_token": self.__refresh_token},
            headers={"X-Client-Version": __version__},
        )

        if not response.ok:
            logger.error(f"Authentication failed: {response.text}")
            sys.exit(1)

        # TODO: Technically the /login endpoint can return a challenge for you
        #       to change your password.
        #       Now that we have a web based login flow we should force them to
        #       do a locust --cloud --login if we get that.

        id_token = response.json()["cognito_client_id_token"]
        id_token_expires = response.json()["id_token_expires"]
        self.__id_token_expires = id_token_expires - 60  # Refresh 1 minute before expiry
        self.headers["Authorization"] = f"Bearer {id_token}"

        if not self.non_interactive:
            config = read_cloud_config()
            config.id_token = id_token
            config.id_token_expires = id_token_expires
            write_cloud_config(config)

    def request(self, method, url, *args, **kwargs) -> requests.Response:
        self.__ensure_valid_authorization_header()
        return super().request(method, f"{self.api_url}{url}", *args, **kwargs)



import importlib.metadata
import json
import os
import pathlib
from dataclasses import dataclass

import platformdirs

__version__ = importlib.metadata.version("locust-cloud")


VALID_REGIONS = ["us-east-1", "eu-north-1"]
CLOUD_CONF_FILE = pathlib.Path(platformdirs.user_config_dir(appname="locust-cloud")) / "config"


@dataclass
class CloudConfig:
    id_token: str | None = None
    user_sub_id: str | None = None
    refresh_token: str | None = None
    refresh_token_expires: int = 0
    region: str | None = None
    id_token_expires: int = 0


def get_api_url(region):
    return os.environ.get("LOCUSTCLOUD_DEPLOYER_URL", f"https://api.{region}.locust.cloud/1")


def read_cloud_config() -> CloudConfig:
    if CLOUD_CONF_FILE.exists():
        with open(CLOUD_CONF_FILE) as f:
            return CloudConfig(**json.load(f))

    return CloudConfig()


def write_cloud_config(config: CloudConfig) -> None:
    CLOUD_CONF_FILE.parent.mkdir(parents=True, exist_ok=True)

    with open(CLOUD_CONF_FILE, "w") as f:
        json.dump(config.__dict__, f)


def delete_cloud_config() -> None:
    if CLOUD_CONF_FILE.exists():
        CLOUD_CONF_FILE.unlink()
